
import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.special import gammaln, psi, factorial
from scipy.optimize import fmin_l_bfgs_b as optim
from statsmodels.genmod.generalized_linear_model import GLM
from statsmodels.genmod.families import NegativeBinomial

from pydgea.prework import obtain_contrast_vec


## I referred to the code in the link below :
## https://github.com/gokceneraslan/fit_nbinom

## GLM
def fit_glm(counts, design_matrix) :
    
    """
    Parameters
    ----------
    counts : pandas.Series
        Raw counts for a given gene.
    
    design_matrix : pandas.DataFrame
        Design matrix.
    
    Returns
    -------
    statsmodels.GLMResultsWrapper
    """
    
    exog = design_matrix.copy()
    endog = counts.copy()
        
    model = GLM(endog=endog, exog=exog, family=NegativeBinomial())
    res = model.fit()
        
    return res


## Fit negative binomial distribution with the maximum likelihood estimation. 
def ll_nbinom(params, data):
    
    """
    Parameters
    ----------
    params : list 
        A list in the form of [r, p].
        r : float greater than or equal to 0. The number of success. 
        p : float between 0 and 1. The probability of a single success.
    
    data : numpy.ndarray
        The observed data.
        
    Returns
    -------
    float
        Negative log likelihood of fitted negative binomial model.
    """
    
    infinitesimal = np.finfo(float).eps
    
    r, p = params 
    if r < 0 :
        raise ValueError('r must be positive or 0')
    elif p < 0 or p > 1 :
        raise ValueError('p must be between 0 and 1')
    
    # Refer http://en.wikipedia.org/wiki/Negative_binomial_distribution#Maximum_likelihood_estimation
    N = len(data)
    ll = np.sum(gammaln(data + r)) \
        - np.sum(np.log(factorial(data))) \
        - N*(gammaln(r)) \
        + N*r*np.log(p) \
        + np.sum(data*np.log(1-(p if p < 1 else 1-infinitesimal)))

    return -ll


def ll_nbinom_deriv(params, data) :
    
    """
    Parameters
    ----------
    params : list 
        A list in the form of [r, p].
        r : float greater than or equal to 0. The number of success. 
        p : float between 0 and 1. The probability of a single success.
    
    data : numpy.ndarray
        The observed data.
        
    Returns
    -------
    numpy.ndarray
        The first element is a partial differential value of log likelihood with respect to r. 
        The second element is a partial differential value with respect to p.
    """
    
    infinitesimal = np.finfo(float).eps
    
    r, p = params
    if r < 0 :
        raise ValueError('r must be positive or 0')
    elif p < 0 or p > 1 :
        raise ValueError('p must be between 0 and 1')

    # Refer http://en.wikipedia.org/wiki/Negative_binomial_distribution#Maximum_likelihood_estimation
    N = len(data)
    pderiv = (N*r)/p - np.sum(data)/(1-(p if p < 1 else 1-infinitesimal))
    rderiv = np.sum(psi(data + r)) \
            - N*psi(r) \
            + N*np.log(p)

    return np.array([rderiv, pderiv])


def obtain_mu(counts, design_matrix, size_factors) :
    
    """
    Parameters
    ----------
    counts : pandas.Series
        Raw counts for a given gene.
    
    design_matrix : pandas.DataFrame
        Design matrix.
    
    size_factors : pandas.Series
        Normalization factors.
        
    Returns
    -------
    pandas.Series
        Mean estimates for a given gene.
    """
    
    glm = fit_glm(counts/size_factors, design_matrix)
    
    mu = np.exp(design_matrix @ glm.params)
    mu = mu * size_factors
    
    return mu


def fit_mle_nbinom(counts, design_matrix, size_factors, init_alpha) :
    
    """
    Parameters
    ----------
    counts : pandas.Series
        Raw counts for a given gene.
    
    design_matrix : pandas.DataFrame
        Design matrix.
    
    size_factors : pandas.Series
        Normalization factors.
        
    init_alpha : float
        Initial alpha(= 1/r).
        
    Returns
    -------
    float
        Dispersion estimate for a given gene.
    """

    infinitesimal = np.finfo(float).eps
    
    mu = obtain_mu(counts, design_matrix, size_factors)
    mu = mu.values
    design_matrix = design_matrix.values

    def loss(log_alpha: float) -> float:

        alpha = np.exp(log_alpha) 
        r = 1/alpha
        
        nll = 0 
        for i in range(len(counts)) :
            count = counts[i]
            mean = mu[i] 

            p = 1/(1 + mean*alpha)

            nll += ll_nbinom([r,p], np.array([count]))
            
        weight = mu / (1 + mu * alpha)
        reg = 0.5 * np.linalg.slogdet((design_matrix.T * weight) @ design_matrix)[1]

        return nll + reg

    def dloss(log_alpha: float) -> float:

        alpha = np.exp(log_alpha)  
        r = 1/alpha

        ll_deriv = 0
        for i in range(len(counts)) :
            count = counts[i]
            mean = mu[i] 

            p = 1/(1 + mean*alpha)

            ll_deriv += ll_nbinom_deriv([r,p], np.array([count]))[0]
            
        weight = mu / (1 + mu * alpha)
        d_weight = -(weight**2)
        reg_grad = (0.5 * (np.linalg.inv((design_matrix.T * weight) @ design_matrix)
                * ((design_matrix.T * d_weight) @ design_matrix)).sum()) * alpha

        return alpha * ll_deriv + reg_grad

    min_disp = infinitesimal
    max_disp = 1e+10
    optimres = optim(loss,
                    x0=np.log(init_alpha),
                    fprime=dloss,
                    approx_grad=1,
                    bounds=[(np.log(min_disp), np.log(max_disp))])
    
    return np.exp(optimres[0][0])
    

## Estimate dispersion.
def estimate_dispersions(count_data, design_matrix, size_factors) : 
    
    """
    Parameters
    ----------
    count_data : pandas.DataFrame
        Raw counts.
    
    design_matrix : pandas.DataFrame
        Design matrix.
    
    size_factors : pandas.Series
        Normalization factors.
        
    Returns
    -------
    pandas.DataFrame
        Dispersion estimate for each gene.
    """
    
    genes = count_data.index
    
    disps = pd.DataFrame(index=genes, columns=['dispersion'])
    for gene in genes :
        counts = count_data.loc[gene,:]
        
        mean = np.mean(counts)
        var = np.var(counts)
        init_r = (mean**2)/((var - mean) if var > mean else 10)
        
        init_alpha = 1/init_r
        alpha = fit_mle_nbinom(counts, design_matrix, size_factors, init_alpha)
        disps.loc[gene, 'dispersion'] = alpha
        
    return disps


## Estimate log2 fold change.
def estimate_lfcs(count_data, design_matrix, size_factors) :
    
    """
    Parameters
    ----------
    count_data : pandas.DataFrame
        Raw counts.
    
    design_matrix : pandas.DataFrame
        Design matrix.
    
    size_factors : pandas.Series
        Normalization factors.
        
    Returns
    -------
    pandas.DataFrame
        Log2 fold change estimates for each gene.
    """
    
    genes = count_data.index
    
    levels = design_matrix.columns
    lfcs = pd.DataFrame(index=genes, columns=levels) 
    for gene in genes : 
        counts = count_data.loc[gene,:]
        
        glm = fit_glm(counts/size_factors, design_matrix)
        lfc = glm.params
        
        lfcs.loc[gene,levels] = lfc[levels]
        
    return lfcs
        

def obtain_lfc(lfcs, contrast, design_matrix) : 
    
    """
    Parameters
    ----------
    lfcs : pandas.DataFrame
        Log2 fold change estimates for each gene.
        
    contrast : list
        A list in the form of [factor of interest, treat, control].
        
    design_matrix : pandas.DataFrame
        Design matrix.
        
    Returns
    -------
    pandas.DataFrame
        Log2 fold change estimates between control and treat.
    """

    contrast = obtain_contrast_vec(contrast, design_matrix)
    
    lfc = (lfcs @ contrast).astype('float64')
    lfc = lfc/np.log(2)
    lfc = lfc.to_frame(name='lfc')
    
    return lfc

    
## Wald test
def wald_test(count_data, design_matrix, size_factors, disps, lfcs, contrast) :
    
    """
    Parameters
    ----------
    count_data : pandas.DataFrame
        Raw counts.
    
    design_matrix : pandas.DataFrame
        Design matrix.
    
    size_factors : pandas.Series
        Normalization factors. 
    
    disps : pandas.DataFrame
        Dispersion estimate for each gene.
    
    lfcs : pandas.DataFrame
        Log2 fold change estimates for each gene.
    
    contrast : list
        A list in the form of [factor of interest, treat, control].
        
    Returns
    -------
    pandas.DataFrame
        Estimated p-value for each gene.
    """
    
    genes = count_data.index
    
    contrast = obtain_contrast_vec(contrast, design_matrix)
    
    num_levs = design_matrix.shape[1]
    ridge_factor = np.diag(np.repeat(1e-6, num_levs))
    
    mu = np.exp((design_matrix @ lfcs.T).astype('float64')).T
    mu = mu * size_factors
    weight = mu / (1 + mu.T * disps['dispersion']).T
    
    p_values = pd.DataFrame(index=genes, columns=['pvalue'])
    for gene in genes :
        hessian = (design_matrix.T * weight.loc[gene, :]) @ design_matrix
        cov = np.linalg.inv(hessian.values.astype('float64') + ridge_factor)
        cov = cov @ contrast

        wald_se = np.sqrt(cov.T @ hessian @ cov)
        wald_statistic = contrast @ lfcs.loc[gene, :] / wald_se
        wald_p_value = 2 * norm.sf(np.abs(wald_statistic))

        p_values.loc[gene, 'pvalue'] = wald_p_value
        
    return p_values 
    
    



import pandas as pd
import numpy as np
import itertools
import statsmodels.api as sm

from pydgea.utils import dedupe, remove_all, combine_all


## I referred to the code in the link below :
## https://github.com/owkin/PyDESeq2/blob/main/pydeseq2/utils.py

def check_count_data(count_data) :
    
    """
    Parameters
    ----------
    count_data : pandas.DataFrame
        Raw counts.
    """
    
    if isinstance(count_data, pd.DataFrame) :  
        if count_data.isna().any().any() :
            raise ValueError("Missing values are not allowed.")  
        elif not set(count_data.dtypes) <= {np.dtype('int64'), np.dtype('float64')} :
            raise ValueError("Only numbers are allowed.")
        elif len(set(count_data.columns)) != len(count_data.columns) :
            raise ValueError('Same column names are not allowed.')
        elif len(set(count_data.index)) != len(count_data.index) :
            raise ValueError('Same row names are not allowed.')
    else :
        raise ValueError('count_data must be pandas.DataFrame.')
    
    if ((count_data % 1 != 0).any() | (count_data < 0).any()).any() :
        raise ValueError("Only positive integers(or 0) are alllowed")   
    elif (count_data.sum(axis=0) == 0).any() or (count_data.sum(axis=1) == 0).any() :
        raise ValueError('There is a row or column with a sum of 0.')
        

# after check_count_data
def check_metadata(metadata, count_data) :
    
    """
    Parameters
    ----------
    metadata : pandas.DataFrame
        Sample metadata.
    
    count_data : pandas.DataFrame
        Raw counts.
    """
    
    if isinstance(metadata, pd.DataFrame) : 
        if metadata.isna().any().any() :
            raise ValueError("Missing values are not allowed.")
        elif len(set(metadata.columns)) != len(metadata.columns) :
            raise ValueError('Same column names are not allowed.')
        elif len(set(metadata.index)) != len(metadata.index) :
            raise ValueError('Same row names are not allowed.')
        elif set(metadata.index) != set(count_data.columns) :
            raise ValueError('The samples in metadata and the samples in count_data must be the same.')
    else :
        raise ValueError('metadata must be pandas.DataFrame.')
        
    for colname in metadata.columns :
        if '~' in colname or '+' in colname or ':' in colname :
            raise ValueError("Any column name of metadata must not contain '~', '+', ':'")
            
            
# after check_metadata
def obtain_design_factors(design, metadata) :
    
    """
    Parameters
    ----------
    design : str
        Design formula.
    
    metadata : pandas.DataFrame
        Sample metadata.
        
    Returns
    -------
    list
        Column names of metadata to be used as design factors.
    """
    
    assert design.strip().startswith('~'), 'The design formula must start with tilde(~).'
    
    table = design.maketrans({'~' : ' ', '+' : ' '})
    design_factors = design.translate(table).split(' ')
    design_factors = remove_all(design_factors, '')
    
    if design_factors == [] : 
        raise ValueError('There are no design factors.')
        
    for i in range(len(design_factors)) :
        factor = design_factors[i]
        if ':' in factor :
            inter = factor.split(':')
            inter = [factor.strip() for factor in inter]
            design_factors[i] = tuple(inter)
            if len(inter) != 2 or len(set(inter)) != 2 :
                raise ValueError('An interaction term must contain only two different factors.')

    if len(dedupe(design_factors)) != len(design_factors) :
        raise ValueError('Same design terms are not allowed.')
        
    for factor in design_factors :
        if type(factor) == tuple :
            inter = factor
            factor1, factor2 = inter
            if (factor2, factor1) in design_factors :
                raise ValueError('Same design terms are not allowed')
            elif factor1 not in design_factors or factor2 not in design_factors :
                raise ValueError('There is a factor with interaction that does not belong to the design factors.')
        else :
            if factor not in metadata.columns :
                raise ValueError('There is a design factor that does not belong to columns of metadata.')
    
    return design_factors


# after obtain_design_factors
def obtain_design_matrix(design_factors, metadata) :
    
    """
    Parameters
    ----------
    design_factors : list
        Design factors.
    
    metadata : pandas.DataFrame
        Sample metadata.
        
    Returns
    -------
    pandas.DataFrame
        A Dataframe with experiment design informatiion. 
    """
           
    total =[]
    inters = []
    for factor in design_factors :
        if type(factor) == tuple :
            inters.append(factor)
        else :
            total.append(factor)

    non_inters = total
    for factor in dedupe(combine_all(inters)) :
        non_inters = remove_all(non_inters, factor)
    
    design_matrix = metadata[non_inters]
    design_matrix = pd.get_dummies(design_matrix, dtype='int', drop_first=True)
    
    if inters != [] :
        for inter in inters :
            factor1, factor2 = inter 
            dm_factor1 = pd.get_dummies(metadata[factor1], dtype='int', drop_first=True)
            dm_factor2 = pd.get_dummies(metadata[factor2], dtype='int', drop_first=True)
            dm_inter = pd.DataFrame(index = dm_factor1.index)
            for level1, level2 in itertools.product(dm_factor1.columns, dm_factor2.columns) :
                dm_inter['{}_{}_and_{}_{}'.format(factor1, level1, factor2, level2)] = dm_factor1[level1] * dm_factor2[level2]
            dm_inter = pd.concat([dm_factor1, dm_factor2, dm_inter], axis=1)
            design_matrix = pd.concat([design_matrix, dm_inter], axis=1)
            
    design_matrix = sm.add_constant(design_matrix)
    design_matrix.columns.values[0] = 'intercept'
    
    return design_matrix


def check_contrast(contrast, design_factors, metadata) :
    
    """
    Parameters
    ----------
    contrast : list
        A list in the form of [factor of interest, treat, control].
    
    design_factors : list
        Design factors.
    
    metadata : pandas.DataFrame
        Sample metadata.
    """
    
    assert type(contrast) == list, 'contrast must be a list.'

    if len(contrast) != 3 :
        raise ValueError("contrast must be length of 3.")
    else :
        factor, treat, control = contrast[0], contrast[1], contrast[2]   
        
    levels = dedupe(metadata[factor])
    
    if control == treat :
        raise ValueError("control and treat must be different.")
    elif control not in levels or treat not in levels :
        raise ValueError("control and treat must belong to the levels of factor of interest.")
        

# after check_contrast
def obtain_contrast_vec(contrast, design_matrix) : 
    
    """
    Parameters
    ----------
    contrast : list
        A list in the form of [factor of interest, treat, control].
    
    design_matrix : pandas.DataFrame
        Design matrix.
        
    Returns
    -------
    numpy.ndarray
        Vector encoding the contrast.
    """
    
    factor, treat, control = contrast[0], contrast[1], contrast[2]   
    
    design_levels = design_matrix.columns
    contrast_vec = pd.Series(0, index=design_levels)
    
    control_level = '{}_{}'.format(factor, control)
    treat_level = '{}_{}'.format(factor, treat)
    if control_level in design_levels and treat_level not in design_levels :
        contrast_vec['{}_{}'.format(factor, control)] = -1
    elif treat_level in design_levels and control_level not in design_levels :
        contrast_vec['{}_{}'.format(factor, treat)] = -1
    elif control_level in design_levels and treat_level in design_levels :
        contrast_vec['{}_{}'.format(factor, control)] = -1
        contrast_vec['{}_{}'.format(factor, treat)] = 1
    else :
        raise ValueError('Invalid contrast.')
    
    return np.array(contrast_vec)
    
    



from setuptools import setup, find_packages


setup(
    name="pyDGEA",
    version='0.0.2',
    python_requires=">=3.9.0",
    license="MIT",
    description="A python package that implements differential gene expression analysis.",
    author="Innocent Labs",
    author_email="jakec63738@gmail.com",
    packages=find_packages()
)